import logging, json, os, nre, sequtils, strtabs, tables
import private/util

# Types
#
type
  BuildStatus* = object
    state*, details*: string

  Step* = object
    name*, stepCmd*, workingDir*: string
    artifacts*, cmdInput*, depends*, expectedEnv*: seq[string]
    dontSkip*: bool

  ProjectCfg* = object
    name*: string
    versionCmd*: string
    steps*: Table[string, Step]

  ProjectDef* = object
    cfgFilePath*, defaultBranch*, name*, repo*: string
    envVars*: StringTableRef

  StrawBossConfig* = object
    artifactsRepo*: string
    projects*: seq[ProjectDef]

# internal utils

let nullNode = newJNull()
proc getIfExists(n: JsonNode, key: string): JsonNode =
  result = if n.hasKey(key): n[key]
           else: nullNode

# Configuration parsing code

proc loadStrawBossConfig*(cfgFile: string): StrawBossConfig =
  if not existsFile(cfgFile):
    raiseEx "strawboss config file not found: " & cfgFile

  let jsonCfg = parseFile(cfgFile)

  var projectDefs: seq[ProjectDef] = @[]

  for pJson in jsonCfg.getIfExists("projects").getElems:
    if not pJson.hasKey("name"):
      raiseEx "a project definition is missing the project name"

    if not pJson.hasKey("repo"):
      raiseEx "a project definition is missing the project repo configuration"

    var envVars = newStringTable(modeCaseSensitive)
    for k, v in pJson.getIfExists("envVars").getFields: envVars[k] = v.getStr("")

    projectDefs.add(
      ProjectDef(
        cfgFilePath: pJson.getIfExists("cfgFilePath").getStr("strawboss.json"),
        defaultBranch: pJson.getIfExists("defaultBranch").getStr("master"),
        name: pJson["name"].getStr,
        envVars: envVars,
        repo: pJson["repo"].getStr))

  result = StrawBossConfig(
    artifactsRepo: jsonCfg.getIfExists("artifactsRepo").getStr("artifacts"),
    projects: projectDefs)

proc loadProjectConfig*(cfgFile: string): ProjectCfg =
  if not existsFile(cfgFile):
    raiseEx "project config file not found: " & cfgFile

  let jsonCfg = parseFile(cfgFile)

  if not jsonCfg.hasKey("name"):
    raiseEx "project configuration is missing a name"

  if not jsonCfg.hasKey("steps"):
    raiseEx "project configuration is missing steps definition"

  var steps = initTable[string, Step]()
  for sName, pJson in jsonCfg["steps"].getFields:
    steps[sName] = Step(
      name:         sName,
      workingDir:   pJson.getIfExists("workingDir").getStr("."),
      stepCmd:      pJson.getIfExists("stepCmd").getStr("sh"),
      depends:      pJson.getIfExists("depends").getElems.mapIt(it.getStr),
      artifacts:    pJson.getIfExists("artifacts").getElems.mapIt(it.getStr),
      cmdInput:     pJson.getIfExists("cmdInput").getElems.mapIt(it.getStr),
      expectedEnv:  pJson.getIfExists("expectedEnv").getElems.mapIt(it.getStr),
      dontSkip:     pJson.getIfExists("dontSkip").getStr("false") != "false")

    if steps[sName].stepCmd == "sh" and steps[sName].cmdInput.len == 0:
      warn "Step " & sName & " uses 'sh' as its command but has no cmdInput."

  result = ProjectCfg(
    name: jsonCfg["name"].getStr,
    versionCmd: jsonCfg.getIfExists("versionCmd").getStr("git describe --tags --always"),
    steps: steps)

proc loadBuildStatus*(statusFile: string): BuildStatus =
  if not existsFile(statusFile): raiseEx "status file not found: " & statusFile
  let jsonObj = parseFile(statusFile)

  if not jsonObj.hasKey("state"):
    raiseEx "project status is missing the 'state' field"

  result = BuildStatus(
    state: jsonObj["state"].getStr,
    details: jsonObj.getIfExists("details").getStr("") )

proc `%`*(s: BuildStatus): JsonNode =
  result = %* {
    "state": s.state,
    "details": s.details
  }

proc `$`*(s: BuildStatus): string =
  result =pretty(%s)
