import logging, nre, os, osproc, sequtils, streams, strtabs, strutils, tables, tempfile

import private/util
import configuration
from posix import fork, Pid

type
  Workspace* = ref object
    artifactsRepo*: string
    env*: StringTableRef
    workingDir*: string
    project*: ProjectCfg

  RunSummary* = object
    project*: ProjectCfg
    step*: Step
    buildVersion*, statusFile*: string
    workerPid*: Pid

  RunRequest* = object
    projectName*, stepName*, buildRef*, workspaceDir*: string
    forceRebuild*: bool

proc resolveEnvVars(line: string, env: StringTableRef): string =
  result = line
  for found in line.findAll(re"\$\w+|\$\{[^}]+\}"):
    let key = if found[1] == '{': found[2..^2] else: found[1..^1]
    if env.hasKey(key): result = result.replace(found, env[key])

proc combineProcMsgHandlers(a, b: HandleProcMsgCB): HandleProcMsgCB =
  if a == nil: result = b
  elif b == nil: result = a
  else:
    result = proc(cmd: string, outMsg, errMsg: TaintedString): void =
      a(cmd, outMsg, errMsg)
      b(cmd, outMsg, errMsg)

proc setupProjectForWork(projectDef: ProjectDef, buildRef, artifactsRepo: string,
                         outputHandler: HandleProcMsgCB = nil): Workspace =

  outputHandler.sendMsg "Setting up to do work for '" & projectDef.name &
                        "' at ref " & buildRef & "."

  var env = loadEnv()
  env["GIT_DIR"] = ".git"

  # Create a temp directory that we'll work in
  let projDir = mkdtemp()
  debug "Workspace for '" & projectDef.name & ":  " & projDir
  assert projDir.isAbsolute

  # Clone the project into the $temp/repo directory
  let cloneResult = exec("git", projDir, ["clone", projectDef.repo, "repo"],
    env, {poUsePath}, outputHandler)

  if cloneResult.exitCode != 0:
    removeDir(projDir)
    raiseEx "unable to clone repo for '" & projectDef.name & "'"

  # Checkout the requested ref
  let checkoutResult = exec("git", projDir & "/repo", ["checkout", buildRef],
    env, {poUsePath}, outputHandler)

  if checkoutResult.exitCode != 0:
    removeDir(projDir)
    raiseEx "unable to checkout ref " & buildRef & " for '" & projectDef.name & "'"

  # Find the strawboss project configuration
  let projCfgFile = projDir & "/repo/" & projectDef.cfgFilePath
  if not existsFile(projCfgFile):
    removeDir(projDir)
    raiseEx "Cannot find strawboss project configuration in the project " &
      "repo (expected at '" & projectDef.cfgFilePath & "')."

  let projectCfg = loadProjectConfig(projCfgFile)
  result = Workspace(env: env, workingDir: projDir, project: projectCfg,
                     artifactsRepo: artifactsRepo)

  # Merge in the project-defined env vars
  for k, v in projectDef.envVars: result.env[k] = v

  # Get the build version
  let versionProc = startProcess(
    projectCfg.versionCmd,      # command
    projDir & "/repo",          # working dir
    [],                         # args
    result.env,                 # environment
    {poUsePath, poEvalCommand}) # options

  let versionResult = waitForWithOutput(versionProc, outputHandler,
                                        projectCfg.versionCmd)

  if versionResult.exitCode != 0:
    removeDir(projDir)
    raiseEx "Version command (" & projectCfg.versionCmd & ") returned non-zero exit code."

  outputHandler.sendMsg "Building version " & versionResult.output.strip
  result.env["VERSION"] = versionResult.output.strip

proc runStep*(step: Step, wksp: Workspace,
              givenOutputHandler: HandleProcMsgCB = nil): void =
  let SB_EXPECTED_VARS = ["VERSION"]
  let stepArtifactDir = wksp.artifactsRepo & "/" & wksp.project.name & "/" &
                        step.name & "/" & wksp.env["VERSION"]
  let statusFile = stepArtifactDir & ".status.json"

  if not existsDir(stepArtifactDir): createDir(stepArtifactDir)

  # Have we tried to build this before and are we caching the results?
  if existsFile(statusFile) and not step.dontSkip:
    let status = loadBuildStatus(statusFile)

    # If we succeeded last time, no need to rebuild
    if status.state == "complete":
      givenOutputHandler.sendMsg "Skipping step '" & step.name & "' for version '" &
        wksp.env["VERSION"] & "': already completed."
      return
    else:
      givenOutputHandler.sendMsg "Rebuilding failed step '" & step.name & "' for version '" &
        wksp.env["VERSION"] & "'."

  givenOutputHandler.sendMsg "Running step '" & step.name & "' for " & wksp.project.name
  writeFile(statusFile, $BuildStatus(state: "running", details: ""))

  var stdoutLogFile, stderrLogFile: File

  try:

    var outputHandler: HandleProcMsgCB

    # Make sure we log output to the stdout and sterr log files
    if not (open(stdoutLogFile, stepArtifactDir & "/stdout.log", fmWrite) and
            open(stderrLogFile, stepArtifactDir & "/stderr.log", fmWrite)):
      givenOutputHandler.sendMsg nil, "Failed to open log files for STDOUT and STDERR."
      outputHandler = givenOutputHandler
      if stdoutLogFile != nil: close(stdoutLogFile)
      if stderrLogFile != nil: close(stderrLogFile)
    else:
      outputHandler = combineProcMsgHandlers(
        givenOutputHandler,
        makeProcMsgHandler(stdoutLogFile, stderrLogFile))

    # Ensure all expected environment variables are present.
    for k in (step.expectedEnv & @SB_EXPECTED_VARS):
      if not wksp.env.hasKey(k):
        debug "workspace.env = " & $(wksp.env)
        raiseEx "step " & step.name & " failed: missing required env variable: " & k

    # Ensure that artifacts in steps we depend on are present
    # TODO: detect circular-references in dependency trees.
    for dep in step.depends:
      if not wksp.project.steps.hasKey(dep):
        raiseEx step.name & " depends on " & dep &
                " but there is no step named " & dep
      let depStep = wksp.project.steps[dep]

      # Run that step (may get skipped)
      let depDir = wksp.artifactsRepo & "/" & wksp.project.name & "/" &
                   dep & "/" & wksp.env["VERSION"]

      runStep(depStep, wksp)

      # Add the artifacts directory for the dependent step to our env so that
      # further steps can reference it via $<stepname>_DIR
      echo "FP: " & depDir
      wksp.env[depStep.name & "_DIR"] = depDir

    # Run the step command, piping in cmdInput
    outputHandler.sendMsg step.name & ": starting stepCmd: " & step.stepCmd
    let cmdProc = startProcess(step.stepCmd,
      wksp.workingDir & "/repo/" & step.workingDir,
      [], wksp.env, {poUsePath, poEvalCommand})

    let cmdInStream = inputStream(cmdProc)

    # Replace env variables in step cmdInput as we pipe it in
    for line in step.cmdInput: cmdInStream.writeLine(line.resolveEnvVars(wksp.env))
    cmdInStream.flush()
    cmdInStream.close()
    
    let cmdResult = waitForWithOutput(cmdProc, outputHandler, step.stepCmd)

    if cmdResult.exitCode != 0:
      raiseEx "step " & step.name & " failed: step command returned non-zero exit code"

    # Gather the output artifacts (if we have any)
    if step.artifacts.len > 0:
      for a in step.artifacts:
        let artifactPath = a.resolveEnvVars(wksp.env)
        let artifactName = artifactPath[(artifactPath.rfind("/")+1)..^1]
        try:
          copyFile(wksp.workingDir & "/repo/" & step.workingDir & "/" & artifactPath,
                   stepArtifactDir & "/" & artifactName)
        except:
          raiseEx "step " & step.name & " failed: unable to copy artifact " &
            artifactPath & ":\n" & getCurrentExceptionMsg()

    writeFile(statusFile, $BuildStatus(state: "complete", details: ""))

  except:
    writeFile(statusFile, $BuildStatus(
      state: "failed",
      details: getCurrentExceptionMsg()))

  finally:
    if stdoutLogFile != nil: close(stdoutLogFile)
    if stderrLogFile != nil: close(stderrLogFile)

proc runStep*(cfg: StrawBossConfig, req: RunRequest,
              outputHandler: HandleProcMsgCB = nil): RunSummary =

  if not existsDir(req.workspaceDir): createDir(req.workspaceDir)
  let statusFile = req.workspaceDir & "/" & "status.json"

  try:
    writeFile(statusFile, $BuildStatus(
      state: "setup",
      details: "Preparing working environment."))

    # Find the project definition
    let matching = cfg.projects.filterIt(it.name == req.projectName)
    if matching.len == 0: raiseEx "no such project: " & req.projectName
    elif matching.len > 1: raiseEx "more than one project named : " & req.projectName

    let projectDef = matching[0]

    # Find the commit reference we're building
    let foundBuildRef =
      if req.buildRef != nil and req.buildRef.len > 0: req.buildRef
      else: projectDef.defaultBranch

    let wksp = setupProjectForWork(projectDef, foundBuildRef, cfg.artifactsRepo,
                                   outputHandler)

    # Find the step
    if not wksp.project.steps.hasKey(req.stepName):
      raiseEx "no step name '" & req.stepName & "' for " & req.projectName

    var step = wksp.project.steps[req.stepName]

    # Enfore forceRebuild
    if req.forceRebuild: step.dontSkip = true

    result = RunSummary(
      project: wksp.project,
      step: step,
      buildVersion: wksp.env["VERSION"],
      statusFile: wksp.artifactsRepo & "/" & wksp.project.name & "/" & step.name &
                                       "/" & wksp.env["VERSION"] & ".status.json")
    if req.async:
      let pid = fork()
      if pid == 0: runStep(step, wksp, outputHandler) # if we are the child
      else: result.workerPid = pid                    # if we are the parent

    else: runStep(step, wksp, outputHandler)
